//--------------------------------------------------------------------------------
// Project LibAlign
//
// Copyright (C) 2000 Andreas Heger All rights reserved
//
// Author: Andreas Heger <heger@ebi.ac.uk>
//
// $Id: ImplMatrix.cpp,v 1.1.1.1 2002/07/08 21:20:17 heger Exp $
//--------------------------------------------------------------------------------    


#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <cstring> // for memcpy
#include "phylolib.h"
#include "ImplMatrix.h"
#include "PhyloException.h"

using namespace std;

namespace PhyloLib {

//define swap_temp for this macro to work
#define SWAP(x,y) { swap_temp = x; x = y; y = swap_temp; }

//---------------------------------------------------------< constructors and destructors >--------------------------------------

ImplMatrix::ImplMatrix() : mWidth(0), mSize(0), mMatrix( NULL) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplMatrix::ImplMatrix()" << endl;
#endif
}
  
ImplMatrix::ImplMatrix( TYPE_MATRIX_SIZE width, TYPE_MATRIX_VALUE default_value) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplMatrix::ImplMatrix(TYPE_MATRIX_SIZE width, TYPE_MATRIX_VALUE default_value )" << endl;
#endif

  allocateMemory();

  TYPE_MATRIX_SIZE i;
  for (i = 0; i < mSize; i++) 
    mMatrix[i] = default_value;
}

ImplMatrix::ImplMatrix( TYPE_MATRIX_SIZE width, TYPE_MATRIX_VALUE * source) {
  mWidth  = width;
  mMatrix = source;

  calculateSize();
}


ImplMatrix::~ImplMatrix () {
  freeMemory();
}


ImplMatrix::ImplMatrix (const ImplMatrix & src ) : mWidth( src.mWidth ), mSize( src.mSize ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplMatrix::Matrix( const Matrix & src )" << endl;
#endif
  
  // have to allocate memory myself, because virtual functions do not work inside constructors
  // so calling CalculateSize() is not an option

  mMatrix = new TYPE_MATRIX_VALUE[ mSize ];

  if (!mMatrix)
      throw PhyloException("Out of memory in ImplMatrix::AllocateMemory");
  
  memcpy( mMatrix, src.mMatrix, mSize * sizeof( TYPE_MATRIX_VALUE) );
  
}


//--------------------------------------------------------------------------------------------
TYPE_MATRIX_VALUE ImplMatrix::getMinimum( COORD & x) const { 

  
  TYPE_MATRIX_VALUE min = 999999;

  TYPE_MATRIX_SIZE i;
  TYPE_MATRIX_SIZE best_index = 0;

  for (i = 0; i < mSize; i++) {
    if (mMatrix[i] < min) {
      min = mMatrix[i];
      best_index = i;
    }
  }

  x.row = getRow( best_index );
  x.col = getColumn( best_index );

  return min;
}

//--------------------------------------------------------------------------------------------
TYPE_MATRIX_VALUE ImplMatrix::getMinimum() const {
  COORD x;
  
  return getMinimum(x);
}

//-------------------------------------------------------------
void ImplMatrix::setWidth(TYPE_MATRIX_SIZE width) {
    mWidth = width;
    allocateMemory();
}

//--------------------------------------------------------------------------------------------
TYPE_MATRIX_VALUE ImplMatrix::getMaximum( COORD & x) const { 
  
  TYPE_MATRIX_VALUE max = -999999;
  
  TYPE_MATRIX_SIZE i;
  TYPE_MATRIX_SIZE best_index = 0;

  for (i = 0; i < mSize; i++) {
    if (mMatrix[i] > max)
      {
	max = mMatrix[i];
	best_index = i;
      }
  }

  x.row = getRow( best_index );
  x.col = getColumn( best_index );
  return max;
}

TYPE_MATRIX_VALUE ImplMatrix::getMaximum() const {
  COORD x;
  
  return getMaximum(x);
}


//--------------------------------------------------------------------------------------------
void ImplMatrix::allocateMemory() {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplMatrix::allocateMemory()" << endl;
#endif
  
  // clear old memory, be careful!, use copy constructor to copy matrices
  TYPE_MATRIX_SIZE saved_width = mWidth;
  freeMemory();
  mWidth = saved_width;

#ifdef DEBUG
  cout << "Allocating " << mSize << " bytes for amatrix of width " << mWidth << endl;
#endif

  calculateSize();
  mMatrix = new TYPE_MATRIX_VALUE[ mSize ];

  if (!mMatrix)
      throw PhyloException("Out of memory in ImplMatrix::AllocateMemory");
  
}

//--------------------------------------------------------------------------------------------
void ImplMatrix::freeMemory() {

  if (mMatrix != NULL) 
    delete [] mMatrix;
  
  mMatrix = NULL;
  mWidth  = 0;
  mSize   = 0;
  
}

//-------------------------------------------------------------
void ImplMatrix::Shrink() {
  //!! to be implemented
}

//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrix::getWidth() const {
    return mWidth;
}

//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrix::getSize() const {
    return mSize;
}

//-------------------------------------------------------------
TYPE_MATRIX_VALUE ImplMatrix::operator()( TYPE_MATRIX_SIZE row, TYPE_MATRIX_SIZE col) const {

#ifdef DEBUG
  if (verbose > LL5) { 
    cout << "TYPE_MATRIX_VALUE ImplMatrix::operator()( TYPE_MATRIX_SIZE, TYPE_MATRIX_SIZE) const " << endl;
    cout << "Looking up " << row << " " << col << endl;
    cout << "Index is " << getIndex( row, col ) << endl;
  }
#endif

    return mMatrix[getIndex( row, col)];
}    

//-------------------------------------------------------------
TYPE_MATRIX_VALUE ImplMatrix::getElement( TYPE_MATRIX_SIZE row, TYPE_MATRIX_SIZE col) const {
    return mMatrix[getIndex( row, col)];
}    

//-------------------------------------------------------------
TYPE_MATRIX_VALUE & ImplMatrix::operator()( TYPE_MATRIX_SIZE row, TYPE_MATRIX_SIZE col) {
  return mMatrix[getIndex(row,col)];
}    
//-------------------------------------------------------------
void ImplMatrix::setElement( TYPE_MATRIX_SIZE row, TYPE_MATRIX_SIZE col, TYPE_MATRIX_VALUE value) {
  mMatrix[getIndex(row,col)] = value;
}    

//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrix::getIndex( TYPE_MATRIX_SIZE row, TYPE_MATRIX_SIZE col) const {
    return (row * mWidth + col);
}

//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrix::getRow( TYPE_MATRIX_SIZE index ) const {
    return ( (TYPE_MATRIX_SIZE)(index / mWidth)  ); 
}
  
//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrix::getColumn( TYPE_MATRIX_SIZE index ) const {
    return ( index % mWidth );
}

//-------------------------------------------------------------
void ImplMatrix::calculateSize() {
#ifdef DEBUG
    if (verbose > LL3) cout << "void ImplMatrix::calculateSize()" << endl;
#endif

  mSize = mWidth * mWidth;
}

//-------------------------------------------------------------
void ImplMatrix::Swap( TYPE_MATRIX_SIZE col_1, TYPE_MATRIX_SIZE col_2 ) {
    //!! to be implemented, not finished
    TYPE_MATRIX_VALUE swap_temp;	// for swapping indices
  
    if (col_1 == col_2) return;

    TYPE_MATRIX_SIZE i;

    for (i = 0; i < mWidth; i++) {
	SWAP( mMatrix[getIndex(i, col_1)], mMatrix[getIndex(i, col_2)]);
	SWAP( mMatrix[getIndex(col_1, i)], mMatrix[getIndex(col_2, i)]);
  }
    
}
  
//---------------------------------------------------------< Input/Output routines >---------------------------------------------
void ImplMatrix::Write( std::ostream & output ) const {
  // write so that output can be used in phylip
  
  cout << " " << mWidth << " " << mSize << endl;
  
  TYPE_MATRIX_SIZE i, j;
  
  for (i = 0; i < mWidth; i++) {
    cout << i << "\t";
    for (j = 0; j < mWidth; j++) 
      cout << setw(10) << setprecision(4) << operator()( i, j) << " ";
    cout << endl;
  }
}

//---------------------------------------------------------< Input/Output routines >---------------------------------------------
void ImplMatrix::Read( std::istream & input ) const {
}

//--------------------------------------------------------------------------------------------------------------------------------


} // namespace AlignLib
