//--------------------------------------------------------------------------------
// Project LibAlign
//
// Copyright (C) 2000 Andreas Heger All rights reserved
//
// Author: Andreas Heger <heger@ebi.ac.uk>
//
// $Id: ImplMatrixSymmetric.cpp,v 1.1.1.1 2002/07/08 21:20:17 heger Exp $
//--------------------------------------------------------------------------------    


#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include "phylolib.h"
#include "ImplMatrixSymmetric.h"
#include "PhyloException.h"

using namespace std;

namespace PhyloLib {

Matrix * makeMatrixSymmetric( TYPE_MATRIX_SIZE size, TYPE_MATRIX_VALUE default_value) {
    return new ImplMatrixSymmetric( size, default_value );
}

//---------------------------------------------------------< constructors and destructors >--------------------------------------

ImplMatrixSymmetric::ImplMatrixSymmetric() : ImplMatrix() {
}

ImplMatrixSymmetric::~ImplMatrixSymmetric () {
}

  // not calling the base class constructor is dangerous!!
ImplMatrixSymmetric::ImplMatrixSymmetric (TYPE_MATRIX_SIZE width, TYPE_MATRIX_VALUE default_value ) : ImplMatrix() {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplMatrixSymmetric::ImplMatrixSymmetric (TYPE_MATRIX_SIZE width, TYPE_MATRIX_VALUE default_value )" << endl;
#endif

  mWidth  = width;
  mSize   = mWidth * (mWidth - 1) / 2;

#ifdef DEBUG
  cout << "Allocating " << mSize << " bytes for amatrix of width " << mWidth << endl;
#endif

  mMatrix = new TYPE_MATRIX_VALUE[ mSize ];
  
  TYPE_MATRIX_SIZE i;
  for (i = 0; i < mSize; i++)
    mMatrix[i] = default_value;
}

ImplMatrixSymmetric::ImplMatrixSymmetric( const ImplMatrixSymmetric & src) : ImplMatrix( src ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "void ImplMatrixSymmetric::ImplMatrixSymmetric( const ImplMatrixSymmetric &)" << endl;
#endif

    //!! to be tightened up. Do not copy twice, otherwise memory leak!!!
    // since virtual functions do not work in constructors, I have to everything myself
    
    mWidth = src.getWidth();
    const TYPE_MATRIX_VALUE * matrix = src.mMatrix;
    
    mSize = mWidth * (mWidth - 1) / 2;
    mMatrix = new TYPE_MATRIX_VALUE[ mSize ];

    if (!mMatrix)
	throw PhyloException("Out of memory in ImplMatrixSymmetric");

    TYPE_MATRIX_SIZE i, j;
    TYPE_MATRIX_SIZE index = 0;
    
    for (i = 1; i < mWidth; i++)			// iterate through rows
	for (j = 0; j < i; j++) 			// iterate through columns
	    mMatrix[index++] = matrix[src.getIndex(i,j)];
}

//-------------------------------------------------------------
void ImplMatrixSymmetric::Shrink() {
#ifdef DEBUG
    if (verbose > LL3) cout << "void ImplMatrixSymmetric::Shrink()" << endl;
#endif

  /* simply decrease the width of the matrix */
  mWidth --;
  mSize -= mWidth;

#ifdef DEBUG
  if (verbose > LL3) cout << "New matrix size: " << mWidth << "(" << mSize << ")" << endl;
#endif
}

//-------------------------------------------------------------
void ImplMatrixSymmetric::Swap( TYPE_MATRIX_SIZE row_1, TYPE_MATRIX_SIZE row_2 ) {
  /*
	-			-
	y-			x-
	-y-			-x-
	-y--		->	-x--
	xOxx-			yyyy-
	-y--x-			-x--y-
	-y--x--			-x--y--
	-y--x---		-x--y---
   i.e. two columns have to skipped:
   1. col = row_1
   2. col = row_2
  */
  
  TYPE_MATRIX_VALUE t;
  
#define SWAP( x, y) { t = mMatrix[x], mMatrix[x] = mMatrix[y], mMatrix[y] = t;}

  TYPE_MATRIX_SIZE i;

  if (row_2 < row_1) {
    i = row_1; row_1 = row_2; row_2 = i;
  }

  for (i = 0; i < row_1; i++)
    SWAP( getIndex(row_1,i), getIndex( row_2,i));
  
  for (i = row_1 + 1; i < row_2; i++) 
    SWAP( getIndex(row_1,i), getIndex( row_2,i));

  for (i = row_2 + 1; i < mWidth; i++)
    SWAP( getIndex(row_1,i), getIndex( row_2,i));
  
#undef SWAP
    
}

//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrixSymmetric::getIndex( TYPE_MATRIX_SIZE row, TYPE_MATRIX_SIZE col) const {
  int x;
 
  if (row == col)
    return NO_INDEX;           // the diagonal is not part of the matrix
 
  if (row > col)
    x = row * ( row - 1) / 2 + col;
  else
    x = col * ( col - 1) / 2 + row;
  
  return x;
}
         
//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrixSymmetric::getRow( TYPE_MATRIX_SIZE index ) const {
  
  TYPE_MATRIX_SIZE row = 1;
  TYPE_MATRIX_SIZE row_index = 0;
  
  while (index >= row_index)
    row_index += row++;
  
  return (row - 1);
}


//-------------------------------------------------------------
TYPE_MATRIX_SIZE ImplMatrixSymmetric::getColumn( TYPE_MATRIX_SIZE index ) const {
  // computationally inefficient, maybe use cache
  return (index - getIndex(getRow(index), 0));	
}

//-------------------------------------------------------------
void ImplMatrixSymmetric::calculateSize() {
#ifdef DEBUG
    if (verbose > LL3) cout << "void ImplMatrixSymmetric::calculateSize()" << endl;
#endif

  mSize = mWidth * (mWidth - 1) / 2;
}

//--------------------------------------------------------------------------------------------------------------------------------


} // namespace PhyloLib
