//--------------------------------------------------------------------------------
// Project PhyloLib
//
// Copyright (C) 2000 Andreas Heger All rights reserved
//
// Author: Andreas Heger <heger@ebi.ac.uk>
//
// $Id: ImplTree.cpp,v 1.3 2004/06/02 12:14:34 aheger Exp $
//--------------------------------------------------------------------------------

#include <iostream>
#include <iomanip>
#include <string>
#include <utility>
#include <stack>
#include "ImplTree.h"
#include "PhyloException.h"

using namespace std;

namespace PhyloLib {


//-----------------------------< factory functions >----------------------------------------------------
Tree * makeTree( TYPE_NODE num_leaves) { return new ImplTree( num_leaves); }

std::ostream & operator<<( std::ostream & output, const NODE_INFO & src) {
  output << src.mParent << " " << src.mLeftChild << " " << src.mRightChild << " " 
	 << src.mNumChildren << " " << src.mWeight << " " << src.mHeight;

  return output;
}

//---------------------------------------------------------< constructors and destructors >--------------------------------------
ImplTree::ImplTree () : mNumLeaves(0), mTree(NULL) {
}

ImplTree::ImplTree( size_t num_leaves) : 
  mNumLeaves( num_leaves ), mCurrentNode(0), mTree(NULL) {
  
  if (num_leaves > 0) 
    setNumLeaves( num_leaves );
}
		       
ImplTree::~ImplTree () {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplTree::~ImplTree()" << endl;
#endif
  delete [] mTree;
}

ImplTree::ImplTree (const ImplTree & src ) : 
  mNumLeaves ( src.mNumLeaves ), mCurrentNode( src.mCurrentNode ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplTree::ImplTree( const ImplTree & )" << endl;
#endif

  //!! to be implemented
}

//-------------------------------------------------------< accessors >----------------------------------------
TYPE_NODE ImplTree::getNumLeaves() const {
  return mNumLeaves;
}    

//-------------------------------------------------------< accessors >----------------------------------------
void ImplTree::setNumLeaves(unsigned int num_leaves) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplTree::setNumLeaves(unsigned int num_leaves)" << endl;
#endif

  assert(num_leaves != 0);
  delete [] mTree;
  mNumLeaves = num_leaves;
  mTree = new NODE_INFO[num_leaves * 2 - 1];
  mCurrentNode = mNumLeaves;
  recordLeaves();
}    

TYPE_NODE ImplTree::getNoNode()  const {
  return NO_NODE;
}
  
TYPE_NODE ImplTree::getRoot()  const {
  return mCurrentNode - 1;
}

TYPE_NODE ImplTree::getLeftChild( TYPE_NODE node) const {
  return mTree[node].mLeftChild;
}

TYPE_NODE ImplTree::getRightChild( TYPE_NODE node) const {
  return mTree[node].mRightChild;
}

TYPE_NODE ImplTree::getParent( TYPE_NODE node) const {
  return mTree[node].mParent;
}

TYPE_NODE ImplTree::getNumChildren( TYPE_NODE node) const {
  return mTree[node].mNumChildren;
}

TYPE_HEIGHT ImplTree::getHeight( TYPE_NODE node) const {
  return mTree[node].mHeight;
}

void ImplTree::setHeight( TYPE_NODE node, TYPE_HEIGHT height) {
  mTree[node].mHeight = height;
}

TYPE_WEIGHT ImplTree::getWeight( TYPE_NODE child, TYPE_NODE parent) const {
    return mTree[child].mWeight;
}

void ImplTree::setWeight( TYPE_NODE child, TYPE_NODE parent, TYPE_WEIGHT weight) {
     mTree[child].mWeight = weight;
}

//-------------------------------------------------------< others >-------------------------------------------
TYPE_NODE ImplTree::findRoot( const TYPE_NODE node) const {
  
  TYPE_NODE n = node;
  while( mTree[n].mParent != NO_NODE)
    n = mTree[n].mParent;
  return n;
}

TYPE_NODE ImplTree::joinNodes( const TYPE_NODE node_1, 
			       const TYPE_NODE node_2,
			       const TYPE_WEIGHT edge_weight_1, 
			       const TYPE_WEIGHT edge_weight_2,
			       const bool map_parents) {
#ifdef DEBUG
    if (verbose > LL3) cout << "void ImplTree::joinNodes()" << endl;
#endif
    
    TYPE_NODE node1, node2;

    if (map_parents) {
      node1 = findRoot( node_1 );
      node2 = findRoot( node_2 );
    } else {
      node1 = node_1;
      node2 = node_2;
    }
    
    mTree[node1].mParent = mCurrentNode;
    mTree[node2].mParent = mCurrentNode;
    
    mTree[node1].mWeight = edge_weight_1;
    mTree[node2].mWeight = edge_weight_2;

    mTree[mCurrentNode].mLeftChild = node1;
    mTree[mCurrentNode].mRightChild = node2;
    mTree[mCurrentNode].mNumChildren = mTree[node1].mNumChildren + 
      mTree[node2].mNumChildren;
    
    return mCurrentNode++;
}

//------------------------------------------------------------------------------------------------------------
void ImplTree::removeRoot() {
#ifdef DEBUG
  if (verbose > LL3) std::cout << "void ImplTree::removeRoot()" << endl;
#endif
}
  
//------------------------------------------------------------------------------------------------------------
TYPE_NODE ImplTree::setRoot( TYPE_NODE node_from, TYPE_NODE node_to, TYPE_WEIGHT weight ) {
#ifdef DEBUG
  if (verbose > LL3)
    std::cout << "void ImplTree::reRoot( TYPE_NODE, TYPE_NODE, TYPE_WEIGHT)" << endl;
#endif
  return getRoot();
}

//----------------------------------------------------------------------------------------
void ImplTree::recordLeaves() {

  // leaves are the first nodes in the graph starting at index 0
  for (TYPE_NODE node = 0; node < getNumLeaves(); node ++) {
    
    mTree[node].mHeight = 0;
    mTree[node].mWeight = 0;
    mTree[node].mNumChildren = 1;
    mTree[node].mParent = NO_NODE;
    mTree[node].mLeftChild = NO_NODE;
    mTree[node].mRightChild = NO_NODE;
  }
}


//---------------------------------> methods for creation of node-lists <--------------------------------------------
//--------------------------------------------------------------------------------------------------------------------
TYPE_NODE_VECTOR * ImplTree::getNodesLeaves() const {

    std::vector<TYPE_NODE> * nodes = new std::vector<TYPE_NODE>(getNumLeaves());

    // by definition, all nodes in the graph with number 0 to getNumLeaves - 1 are leaves
    // there is probably a STL version of this
    for (TYPE_NODE i = 0; i < getNumLeaves(); i++)
	(*nodes)[i] = i;

    return nodes;
}
    
//--------------------------------------------------------------------------------------------------------------------
TYPE_NODE_VECTOR * ImplTree::getNodesDepthFirstVisit() const {

  std::vector<TYPE_NODE> * nodes = new std::vector<TYPE_NODE>;

  std::stack< TYPE_NODE> s;
  
  s.push( getRoot() );
  
  while (!s.empty()) {
    
    TYPE_NODE node = s.top();
    s.pop();
    
    nodes->push_back( node );
    
    TYPE_NODE c;
    if ((c = getLeftChild(node)) != NO_NODE)
      s.push( c );
    if ((c = getRightChild(node)) != NO_NODE)
      s.push( c );
  }
  
  return nodes;
}

//--------------------------------------------------------------------------------------------------------------------
void ImplTree::traversePostOrder( TYPE_NODE node, TYPE_NODE_VECTOR * nodes)  const {
  
  if (node == NO_NODE)
    return;

  traversePostOrder(getLeftChild(node), nodes);
  traversePostOrder(getRightChild(node), nodes);

  nodes->push_back( node );
}


TYPE_NODE_VECTOR * ImplTree::getNodesDepthFirstFinish() const {

  std::vector<TYPE_NODE> * nodes = new std::vector<TYPE_NODE>;

  traversePostOrder( getRoot(), nodes);
  return nodes;

};

//--------------------------------------------------------------------------------------------------------------------
TYPE_NODE_VECTOR * ImplTree::getNodesBreadthFirstVisit() const {
  assert( 0 );
  return NULL;
}

//--------------------------------------------------------------------------------------------------------------------
TYPE_NODE_VECTOR * ImplTree::getNodesBreadthFirstFinish() const {
  assert( 0 );
  return NULL;
}

//---------------------------------------------------------< Input/Output routines >---------------------------------------------
void ImplTree::Write( std::ostream& output ) const {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplTree::Write( std::ostream&) const" << endl;
#endif

  for (TYPE_NODE n =0; n < mCurrentNode; n++) {
    std::cout << n << " " << mTree[n] << endl;
  }

}         




} /* namespace PhyloLib */

