//--------------------------------------------------------------------------------
// Project AlignLib
//
// Copyright (C) 2000 Andreas Heger All rights reserved
//
// Author: Andreas Heger <heger@ebi.ac.uk>
//
// $Id: ImplTreetorDistance.cpp,v 1.1.1.1 2002/07/08 21:20:17 heger Exp $
//--------------------------------------------------------------------------------

#include <iostream>
#include <iomanip>
#include <alignlib/MultipleAlignment.h>

#include "ImplTreetorDistance.h"
#include "Tree.h"
#include "Matrix.h"
#include "HelpersMatrix.h"
#include "Distor.h"
#include "PhyloException.h"

using namespace std;

namespace PhyloLib {

//----------< constructors and destructors >--------------------------------------
ImplTreetorDistance::ImplTreetorDistance ( const Distor * distor ) : 
  ImplTreetor(),
  mDistor( distor ) {
}

ImplTreetorDistance::~ImplTreetorDistance () {
}

ImplTreetorDistance::ImplTreetorDistance (const ImplTreetorDistance & src ) : ImplTreetor( src ) {
}

//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistance::startUp( Tree * tree, const AlignLib::MultipleAlignment * mali) const {

  // create a distance matrix
  if (mali != NULL) {
      assert( mali->getWidth() > 0);
      mWorkMatrix = makeMatrixSymmetric( mali->getWidth());
  } else {
      mWorkMatrix = makeMatrixSymmetric(0);
  }

  // call Distor for multiple alignment to fill matrix with distances
  mDistor->calculateMatrix( mWorkMatrix, mali );

  // save a pointer to the tree
  mTree = tree;
}

//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistance::cleanUp() const {

  /* I have to perform this check, as otherwise I got an error
     under gcc on freeing an empty pointer */
  if (mWorkMatrix != NULL)
    delete mWorkMatrix;
}

//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistance::swapHelpers( TYPE_MATRIX_SIZE cluster_1, TYPE_MATRIX_SIZE cluster_2) const {
}

//------------------------------------------------------------------------------------------------------------------------------
Tree * ImplTreetorDistance::calculateTree( Tree * tree, const AlignLib::MultipleAlignment * mali ) const {
#ifdef DEBUG
  if (verbose > LL3) cout << "Tree *ImplTreetorDistance::calculateTree( Tree *, const AlignLib::MultipleAlignment)" << endl;
#endif

  startUp( tree, mali );

  TYPE_MATRIX_SIZE swap_temp;	// for swapping indices

#define SWAP(x,y) { swap_temp = x; x = y; y = swap_temp; }

  //------------------------------------------------------------------------------------------
  // start up 
  TYPE_MATRIX_SIZE width = mWorkMatrix->getWidth();
  TYPE_MATRIX_SIZE i;
  
  /* in this algorithm I assume that the matrix I use is only a half-matrix using the lower diagonal */
  tree->setNumLeaves( width );					// allocate memory for tree

  mIndices = new TYPE_NODE[width];			        /* allocate array to keep track of indices */

  for (i = 0; i < width; i++) mIndices[i] = i;
  
  //----------------------------------------------------------------------------------------------------
  // Perform hierarchical clustering

  TYPE_MATRIX_SIZE last_row = width - 1;
  
  /* shrink distance matrix, until it contains only a single cluster */
  while (last_row > 0) {
    
#ifdef DEBUG
    if (verbose > LL3) {
      cout << "Last row " << last_row << endl;
      cout << "Work matrix " << endl << *mWorkMatrix << endl;
      
      cout << "----------" << endl;
      for (i = 0; i < width; i++) cout << i << " " << mIndices[i] << endl;
      cout << "----------" << endl;
    }
#endif

    // find minimum distance in matrix 
    /*
	-
	--
	---
	----
	-x---
	------
	-------
	--------
      */
    calculateMinimumDistance();
    TYPE_MATRIX_SIZE min_row = mMinimumCoord.row;
    TYPE_MATRIX_SIZE min_col = mMinimumCoord.col;

#ifdef DEBUG
    if (verbose > LL2) cout << "Joining nodes -> " 
			    << "minimum distance :" << mMinimumValue << " " 
			    << "node 1: " << mIndices[min_row] << " (" << min_row << ") " 
			    << "node 2: " << mIndices[min_col] << " (" << min_col << ") " << endl;
#endif
    
    //------------------------------------------------------------------------------------------------------
    // move rows around, so that the last two joined cluster are in the two last rows

    /*
	-			-
	y-			--
	-y-			---
	-y--		->	----
	xOxx-			-----
	-y--x-			------
	-y--x--			xxxxxxx
	-y--x---		yyyyyyyy
    */

    TYPE_MATRIX_SIZE second_row = last_row - 1;		// second to last row

    // exchange row with last row
    mWorkMatrix->Swap( min_row, last_row);
    SWAP( mIndices[min_row], mIndices[last_row]);
    swapHelpers( min_row, last_row);
    
    // exchange col with second to last row
    mWorkMatrix->Swap( min_col, second_row);
    SWAP( mIndices[min_col], mIndices[second_row]);
    swapHelpers( min_col, second_row);

    //------------------------------------------------------------------------------------------------------
    // join the two nodes in the tree giving the correct edge weights
    TYPE_NODE new_node = joinNodes( second_row, last_row );
    
#ifdef DEBUG
    if (verbose >= LL4) cout << "-> new node " << new_node << endl; 
#endif

    //------------------------------------------------------------------------------------------------------
    // calculate distance to new cluster and put them in second to last row
    updateDistanceMatrix( second_row, last_row);

    /* delete last row, update mIndices, so that the last row now
       contains the number of the new cluster id */
    /*
	-	       	-
	--	       	--
	---	       	---
	----	     ->	----
	-----		-----
	------		------
	xxxxxxx		xxxxxxx
	yyyyyyyy
    */
    
    mWorkMatrix->Shrink();
    mIndices[last_row - 1] = new_node;
    
    last_row = mWorkMatrix->getWidth() - 1;
    
  }    
  
  delete [] mIndices;
  cleanUp();	// cleans up mWorkMatrix

  return mTree;
}

} /* namespace PhyloLib */

