//--------------------------------------------------------------------------------
// Project AlignLib
//
// Copyright (C) 2000 Andreas Heger All rights reserved
//
// Author: Andreas Heger <heger@ebi.ac.uk>
//
// $Id: ImplTreetorDistanceNJ.cpp,v 1.1.1.1 2002/07/08 21:20:17 heger Exp $
//--------------------------------------------------------------------------------

#include <iostream>
#include <iomanip>
#include "ImplTreetorDistanceNJ.h"
#include "Tree.h"
#include "Matrix.h"
#include "Distor.h"
#include "HelpersDistor.h"
#include "PhyloException.h"

using namespace std;

namespace PhyloLib {

//------------------------------------------------------< factory functions >------------------------------------------------
Treetor * makeTreetorDistanceNJ( const Distor * distor ) {
  if (distor == NULL) 
    return new ImplTreetorDistanceNJ( getDefaultDistor());
  else 
    return new ImplTreetorDistanceNJ( distor );
}

//---------------------------------------------------------< constructors and destructors >--------------------------------------
ImplTreetorDistanceNJ::ImplTreetorDistanceNJ ( const Distor * distor) : ImplTreetorDistance( distor), mR (NULL) {
}
		       
ImplTreetorDistanceNJ::~ImplTreetorDistanceNJ () {
}

ImplTreetorDistanceNJ::ImplTreetorDistanceNJ (const ImplTreetorDistanceNJ & src ) : ImplTreetorDistance(src), mR( NULL) {
  //!! to be implemented
}

//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistanceNJ::startUp( Tree * tree, const AlignLib::MultipleAlignment * mali) const {
  
  cleanUp();
  ImplTreetorDistance::startUp( tree, mali);

  TYPE_MATRIX_SIZE width = mWorkMatrix->getWidth();
  TYPE_MATRIX_SIZE width_2 = width - 2;

  mR = new TYPE_MATRIX_VALUE[ width ];
  
  TYPE_MATRIX_SIZE i,k;

  for (i = 0; i < width; i++) {
    mR[i] = 0;
    for (k = 0; k < width; k++) 
      mR[i] += (*mWorkMatrix)( i, k );
    mR[i] /= width_2;
  }
}

//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistanceNJ::cleanUp() const {
  if (mR != NULL) 
    delete [] mR;
  
  mR = NULL;
}


//------------------------------------------------------------------------------------------------------------------------------
TYPE_NODE ImplTreetorDistanceNJ::joinNodes( TYPE_MATRIX_SIZE cluster_1, TYPE_MATRIX_SIZE cluster_2 ) const {
#ifdef DEBUG
  if (verbose > LL3) cout << "TYPE_EDGE ImplTreetorDistanceNJ::joinNodes( TYPE_MATRIX_SIZE, TYPE_MATRIX_SIZE) const" << endl;
#endif

  TYPE_MATRIX_VALUE d_ij = (*mWorkMatrix)( cluster_1, cluster_2 );
  TYPE_MATRIX_VALUE d_ik = (d_ij + mR[cluster_1] - mR[cluster_2]) / 2;
  TYPE_MATRIX_VALUE d_jk = d_ij - d_ik;

#ifdef DEBUG
  if (verbose > LL3) cout << "Joining nodes " <<
		       mIndices[cluster_1] << " (" << d_ik << ") with " << 
		       mIndices[cluster_2] << " (" << d_jk << ");" << endl;
#endif

  return mTree->joinNodes( mIndices[ cluster_1 ], 
			   mIndices[ cluster_2 ],
			   d_ik, 
			   d_jk );
};

//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistanceNJ::swapHelpers( TYPE_MATRIX_SIZE cluster_1, TYPE_MATRIX_SIZE cluster_2) const {
  TYPE_MATRIX_VALUE t;
  
  t = mR[cluster_1];
  mR[cluster_1] = mR[cluster_2];
  mR[cluster_2] = t;

}

//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistanceNJ::calculateMinimumDistance() const {
  
  TYPE_MATRIX_VALUE min = 999999;
  
  TYPE_MATRIX_SIZE row, col;
  TYPE_MATRIX_SIZE best_row = 0;
  TYPE_MATRIX_SIZE best_col = 0;
  TYPE_MATRIX_SIZE width = mWorkMatrix->getWidth();
  TYPE_MATRIX_VALUE d; 

  // iterate through rows and columns, not through indices, since ri and rj have
  // to be calculated, and looking up the row/col for a given index is slow.
  
  for (row = 0; row < width - 1; row++) {
    for (col = row + 1; col < width; col++) {
      if ( (d = (*mWorkMatrix)( row, col ) - (mR[row] + mR[col])) < min ) {
	min = d;
	best_row = row;
	best_col = col;
      }
    }
  }

  mMinimumCoord.row = best_row;
  mMinimumCoord.col = best_col;

  mMinimumValue = min;
}
  
//------------------------------------------------------------------------------------------------------------------------------
void ImplTreetorDistanceNJ::updateDistanceMatrix( TYPE_MATRIX_SIZE cluster_1, TYPE_MATRIX_SIZE cluster_2 ) const {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplTreetorDistanceNJ::updateDistanceMatrix( TYPE_MATRIX_SIZE, TYPE_MATRIX_SIZE)" << endl; 
#endif
  // calculate distance to new cluster and put them in cluster_1
    
  //------------------------------------------------------------------------------------------------------
  // see Durbin et al. for description of algorithm and symbols used here.
  // the update proceeds in three steps
  // 1. mR[s]: multiply by |L| - 2 and subract d_si and d_sj
  // 2. calculate new distance d_sk (k = new cluster, is stored in cluster_1, i.e. i)
  // 3. mR[s]: add d_sk , divide by |L| - 3. 
    
  TYPE_MATRIX_SIZE s;

  TYPE_MATRIX_VALUE new_r = 0;		// mR for new cluster

  TYPE_MATRIX_SIZE last_row = mWorkMatrix->getWidth() - 1;
  TYPE_MATRIX_SIZE num_leaves_2 = last_row - 1;
  TYPE_MATRIX_SIZE num_leaves_3 = last_row - 2;

  if (num_leaves_3 == 0)
    num_leaves_3 = 1;			

#ifdef DEBUG
  if (verbose >= LL4) {
    cout << "R:" << endl;
    for (s = 0; s <= last_row; s++) cout << s << " " << setw(5) << setprecision(2) << mR[s] << endl;
  }
#endif

  // iterate to lastrow. if s = cluster_1 d_ss, etc are 0, so nothing changes.
  TYPE_MATRIX_VALUE d_ij = (*mWorkMatrix)( cluster_1, cluster_2);

  for (s = 0; s < last_row; s++) {
    
    // update mR
    mR[s] =  mR[s] * num_leaves_2 
      - (*mWorkMatrix)( cluster_1, s) 
      - (*mWorkMatrix)( cluster_2, s);

    // calculate new value for d_ms
    TYPE_MATRIX_VALUE d_is = (*mWorkMatrix)( cluster_1, s );
    TYPE_MATRIX_VALUE d_js = (*mWorkMatrix)( cluster_2, s );
    
    TYPE_MATRIX_VALUE new_dist = (d_is + d_js - d_ij) / 2.0;

    (*mWorkMatrix)( cluster_1, s) = new_dist; 
    new_r += new_dist;
    
    // update mR
    mR[s] = (mR[s] +  new_dist) / num_leaves_3;
  }    

  mR[cluster_1] = new_r / num_leaves_3;


}
    

} /* namespace AlignLib */

